@@ -2,6 +2,8 @@
 # Copyright (c) 2011-2013 Raphaël Barrois
 # This code is distributed under the two-clause BSD license.
 
+from __future__ import unicode_literals
+
 """Specific versions of XWorkflows to use with Django."""
 
 from django.db import models
@@ -12,20 +14,12 @@ from django.contrib.contenttypes import models as ct_models
 from django.core import exceptions
 from django.forms import fields
 from django.forms import widgets
-from django.utils.encoding import force_unicode
 from django.utils.translation import ugettext_lazy as _
 
-try:
-    from django.utils import timezone
-    now = timezone.now
-    del timezone
-except ImportError:
-    import datetime
-    now = datetime.datetime.now
-    del datetime
-
 from xworkflows import base
 
+from .compat import force_text, now, python_2_unicode_compatible
+
 
 State = base.State
 AbortTransition = base.AbortTransition
@@ -81,12 +75,12 @@ class StateField(models.Field):
     """Holds the current state of a WorkflowEnabled object."""
 
     default_error_messages = {
-        'invalid': _(u"Choose a valid state."),
-        'wrong_type': _(u"Please enter a valid value (got %r)."),
-        'wrong_workflow': _(u"Please enter a value from the right workflow (got %r)."),
-        'invalid_state': _(u"%s is not a valid state."),
+        'invalid': _("Choose a valid state."),
+        'wrong_type': _("Please enter a valid value (got %r)."),
+        'wrong_workflow': _("Please enter a value from the right workflow (got %r)."),
+        'invalid_state': _("%s is not a valid state."),
     }
-    description = _(u"State")
+    description = _("State")
 
     DEFAULT_MAX_LENGTH = 16
 
@@ -184,14 +178,14 @@ class StateField(models.Field):
         args, kwargs = introspector(self)
 
         state_def = tuple(
-            (str(st.name), unicode(st.title)) for st in self.workflow.states)
-        initial_state_def = self.workflow.initial_state.name
+            (str(st.name), str(st.name)) for st in self.workflow.states)
+        initial_state_def = str(self.workflow.initial_state.name)
 
         workflow = (
             "__import__('xworkflows', globals(), locals()).base.WorkflowMeta("
             "'%(class_name)s', (), "
             "{'states': %(states)r, 'initial_state': %(initial_state)r})" % {
-                'class_name': self.workflow.__class__.__name__,
+                'class_name': str(self.workflow.__class__.__name__),
                 'states': state_def,
                 'initial_state': initial_state_def,
             })
@@ -211,7 +205,7 @@ class WorkflowEnabledMeta(base.WorkflowEnabledMeta, models.base.ModelBase):
         use our custom StateField objects.
         """
         workflows = {}
-        for k, v in attrs.iteritems():
+        for k, v in attrs.items():
             if isinstance(v, StateField):
                 workflows[k] = v
         return workflows
@@ -233,16 +227,30 @@ class WorkflowEnabledMeta(base.WorkflowEnabledMeta, models.base.ModelBase):
         pass
 
 
-class WorkflowEnabled(base.BaseWorkflowEnabled):
+class BaseWorkflowEnabled(base.BaseWorkflowEnabled):
     """Base class for all django models wishing to use a Workflow."""
-    __metaclass__ = WorkflowEnabledMeta
 
     def _get_FIELD_display(self, field):
         if isinstance(field, StateField):
             value = getattr(self, field.attname)
-            return force_unicode(value.title)
+            return force_text(value.title)
         else:
-            return super(WorkflowEnabled, self)._get_FIELD_display(field)
+            return super(BaseWorkflowEnabled, self)._get_FIELD_display(field)
+
+
+# Workaround for metaclasses on python2/3.
+# Equivalent to:
+# Python2
+#
+# class WorkflowEnabled(BaseWorkflowEnabled):
+#     __metaclass__ = WorkflowEnabledMeta
+#
+# Python3
+#
+# class WorkflowEnabled(metaclass=WorkflowEnabledMeta):
+#     pass
+
+WorkflowEnabled = WorkflowEnabledMeta(str('WorkflowEnabled'), (BaseWorkflowEnabled,), {})
 
 
 def get_default_log_model():
@@ -340,6 +348,7 @@ class Workflow(base.Workflow):
             self.db_log(transition, from_state, instance, *args, **kwargs)
 
 
+@python_2_unicode_compatible
 class BaseTransitionLog(models.Model):
     """Abstract model for a minimal database logging setup.
 
@@ -365,19 +374,19 @@ class BaseTransitionLog(models.Model):
     MODIFIED_OBJECT_FIELD = ''
     EXTRA_LOG_ATTRIBUTES = ()
 
-    transition = models.CharField(_(u"transition"), max_length=255,
+    transition = models.CharField(_("transition"), max_length=255,
         db_index=True)
-    from_state = models.CharField(_(u"from state"), max_length=255,
+    from_state = models.CharField(_("from state"), max_length=255,
         db_index=True)
-    to_state = models.CharField(_(u"to state"), max_length=255,
+    to_state = models.CharField(_("to state"), max_length=255,
         db_index=True)
-    timestamp = models.DateTimeField(_(u"performed at"),
+    timestamp = models.DateTimeField(_("performed at"),
         default=now, db_index=True)
 
     class Meta:
         ordering = ('-timestamp', 'transition')
-        verbose_name = _(u'XWorkflow transition log')
-        verbose_name_plural = _(u'XWorkflow transition logs')
+        verbose_name = _('XWorkflow transition log')
+        verbose_name_plural = _('XWorkflow transition logs')
         abstract = True
 
     def get_modified_object(self):
@@ -395,8 +404,8 @@ class BaseTransitionLog(models.Model):
         })
         return cls.objects.create(**kwargs)
 
-    def __unicode__(self):
-        return u'%r: %s -> %s at %s' % (self.get_modified_object(),
+    def __str__(self):
+        return '%r: %s -> %s at %s' % (self.get_modified_object(),
             self.from_state, self.to_state, self.timestamp.isoformat())
 
 
@@ -416,9 +425,9 @@ class GenericTransitionLog(BaseTransitionLog):
     MODIFIED_OBJECT_FIELD = 'modified_object'
 
     content_type = models.ForeignKey(ct_models.ContentType,
-                                     verbose_name=_(u"Content type"),
+                                     verbose_name=_("Content type"),
                                      blank=True, null=True)
-    content_id = models.PositiveIntegerField(_(u"Content id"),
+    content_id = models.PositiveIntegerField(_("Content id"),
         blank=True, null=True, db_index=True)
     modified_object = generic.GenericForeignKey(
             ct_field="content_type",
@@ -426,8 +435,8 @@ class GenericTransitionLog(BaseTransitionLog):
 
     class Meta:
         ordering = ('-timestamp', 'transition')
-        verbose_name = _(u'XWorkflow transition log')
-        verbose_name_plural = _(u'XWorkflow transition logs')
+        verbose_name = _('XWorkflow transition log')
+        verbose_name_plural = _('XWorkflow transition logs')
         abstract = True
 
 
@@ -435,8 +444,8 @@ class BaseLastTransitionLog(BaseTransitionLog):
     """Alternate abstract model holding only the latest transition."""
 
     class Meta:
-        verbose_name = _(u'XWorkflow last transition log')
-        verbose_name_plural = _(u'XWorkflow last transition logs')
+        verbose_name = _('XWorkflow last transition log')
+        verbose_name_plural = _('XWorkflow last transition logs')
         abstract = True
 
     @classmethod
@@ -480,18 +489,18 @@ class GenericLastTransitionLog(BaseLastTransitionLog):
     MODIFIED_OBJECT_FIELD = 'modified_object'
 
     content_type = models.ForeignKey(ct_models.ContentType,
-                                     verbose_name=_(u"Content type"),
+                                     verbose_name=_("Content type"),
                                      related_name='last_transition_logs',
                                      blank=True, null=True)
-    content_id = models.PositiveIntegerField(_(u"Content id"),
+    content_id = models.PositiveIntegerField(_("Content id"),
         blank=True, null=True, db_index=True)
     modified_object = generic.GenericForeignKey(
             ct_field="content_type",
             fk_field="content_id")
 
     class Meta:
-        verbose_name = _(u'XWorkflow last transition log')
-        verbose_name_plural = _(u'XWorkflow last transition logs')
+        verbose_name = _('XWorkflow last transition log')
+        verbose_name_plural = _('XWorkflow last transition logs')
         abstract = True
         unique_together =  ('content_type', 'content_id')
 
